from hypothesis import given, strategies as st, settings
import tensorflow as tf
import keras
import numpy as np
from tf.ctc import edit_distance, accuracy, onehot_logits

@settings(deadline=None)
@given(st.lists(st.lists(st.integers(min_value=1, max_value=64), min_size=1, max_size=64), min_size=1, max_size=8))
def test_perfect_ed(labs: list[list[int]]):
  batch = tf.sparse.from_dense(keras.utils.pad_sequences(labs))
  logits = onehot_logits(batch)
  assert tf.reduce_all(edit_distance(batch, logits) == 0)

@settings(deadline=None)
@given(st.lists(st.lists(st.integers(min_value=1, max_value=64), min_size=1, max_size=64), min_size=1, max_size=8))
def test_perfect_acc(labs: list[list[int]]):
  batch = tf.sparse.from_dense(keras.utils.pad_sequences(labs, padding='post', truncating='post'))
  logits = onehot_logits(batch)
  assert np.isclose(accuracy(batch, logits), 1)


from hypothesis import given, strategies as st, settings
import tensorflow as tf
import numpy as np
from tf.ctc import loss, onehot_logits

@settings(deadline=None)
@given(st.lists(st.integers(min_value=1, max_value=64), min_size=1, max_size=64), st.integers(min_value=0, max_value=12))
def test_zero(lab: list[int], delta: int):
  min_depth = tf.reduce_max(lab)+1
  depth = min_depth + delta
  vlab = tf.sparse.from_dense([lab])
  logits = onehot_logits(vlab, vocabsize=depth)
  assert np.isclose(loss(vlab, logits), 0)


try:
  import pytest
  if __name__ == '__main__':
    import os
    base = os.path.dirname(__file__)
    pytest.main([base, '--verbose'])
except ImportError:
  print('Install testing libs to run tests: `pip install tf-ctc[test]`')


from .loss_ import loss
from .decoding import beam_decode, greedy_decode
from .metrics import accuracy, edit_distance
from .logits import onehot_logits, onehot_logits_sample, mock_logits

__all__ = [
  'loss',
  'beam_decode', 'greedy_decode',
  'accuracy', 'edit_distance',
  'onehot_logits', 'onehot_logits_sample', 'mock_logits'
]


import tensorflow as tf
from jaxtyping import Float, Int
from tf.tools import tf_function

@tf_function
def loss(
  labels: Int[tf.SparseTensor, "batch maxlen"],
  logits: Float[tf.Tensor, "batch maxlen vocab"],
  blank_index: int = 0
) -> Float[tf.Tensor, "batch"]:
  """CTC loss. `blank_index` is the index in the vocabulary corresponding to the blank character"""
  batch = tf.shape(logits)[0]
  maxlen = tf.shape(logits)[1]
  logit_lens = tf.repeat(maxlen, batch)
  return tf.nn.ctc_loss(tf.cast(labels, tf.int32), logits, None, logit_lens, logits_time_major=False, blank_index=blank_index)


import tensorflow as tf
from jaxtyping import Float, Int
from tf.tools import tf_function

@tf_function
def beam_decode_tf(
	logits: Float[tf.Tensor, "batch maxlen vocabsize"],
	beam_width: int = 100, top_paths: int = 1
) -> tuple[list[Int[tf.SparseTensor, "batch maxlen"]], Float[tf.Tensor, "batch 1"]]:
	"""CTC beam search decoding
	- `logits :: FloatTensor[batch, maxlen, vocabsize]`
	- Returns `(top_paths, log_probs)`
		- `top_paths :: list[IntSparseTensor[batch, maxlen]]` where elements correspond to indices in the vocabulary or 0 for blank (so, in the closed interval `[0, vocabsize]`)
		- `log_probs :: list[FloatTensor[batch, 1]]`
		
		Note that `len(top_paths) == len(log_probs) == top_paths`
	"""
	batch = tf.shape(logits)[0] # type: ignore
	maxlen = tf.shape(logits)[1] # type: ignore
	time_major_logits = tf.transpose(logits, perm=[1, 0, 2])
	top_preds, log_probs = tf.nn.ctc_beam_search_decoder(
		inputs=time_major_logits,
		sequence_length=tf.repeat(maxlen, batch),
		beam_width=beam_width, top_paths=top_paths,
	) # type: ignore
	return top_preds, log_probs

@tf_function
def shift_left(logits: tf.Tensor) -> tf.Tensor:
    """Shifts last dimension left by 1: `0 -> n-1, 1 -> 0, 2 -> 1, ...`"""
    vocab_size = tf.shape(logits)[-1] # type: ignore
    index_mapping = tf.concat([
        tf.range(1, vocab_size),  # Shift non-blank indices up by 1
        tf.constant([0], dtype=tf.int32)  # Move blank to the end
    ], axis=0)
    return tf.gather(logits, index_mapping, axis=-1)

@tf_function
def beam_decode(
	logits: Float[tf.Tensor, "batch maxlen vocabsize"],
	beam_width: int = 100, top_paths: int = 1, blank_zero: bool = True
) -> tuple[list[Int[tf.SparseTensor, "batch maxlen"]], Float[tf.Tensor, "batch 1"]]:
	"""
CTC beam search decoding, but blank index can be set to 0
- `logits :: FloatTensor[batch, maxlen, vocabsize]`
- `blank_zero`: if `True` (the default), the blank index is set to 0
- Returns `(top_paths, log_probs)`
- `top_paths :: list[IntSparseTensor[batch, maxlen]]` where elements correspond to indices in the vocabulary or 0 for blank (so, in the closed interval `[0, vocabsize]`)
- `log_probs :: list[FloatTensor[batch, 1]]`
- Note that `len(top_paths) == len(log_probs) == top_paths`


#### How does `blank_zero = True` work?
1. Shift the logit elements left by 1 in the last dimension (`0` goes to `n-1`, `1` to `0`, `2` to `1`, etc.)
2. Run the usual `tf.nn.ctc_beam_decoder`
3. Add 1 to the predicted indices (so, `0 -> 1, 1 -> 2, n-2 -> n-1`. `n-1` was blank thus there can be no `n-1` in the predicted sequences)
	"""
	z = shift_left(logits) if blank_zero else logits
	t, log_probs = beam_decode_tf(z, beam_width, top_paths)
	top_preds: list[tf.SparseTensor] = t # thank you python type linter!
	paths = [s.with_values(s.values+1) for s in top_preds] if blank_zero else top_preds # type: ignore
	return paths, log_probs

@tf_function
def greedy_decode(
    logits: Float[tf.Tensor, "batch maxlen vocabsize"]    
) -> tuple[list[Int[tf.SparseTensor, "batch maxlen"]], Float[tf.Tensor, "batch 1"]]:
	"""CTC greedy decoding
	- `logits :: FloatTensor[batch, maxlen, vocabsize]`
	- Returns `(top_paths, log_probs)`. Just like `beam_decode`, except these are both singleton lists
		- `top_paths[0] :: IntSparseTensor[batch, maxlen]` where elements correspond to indices in the vocabulary or 0 for blank (so, in the closed interval `[0, vocabsize]`)
		- `log_probs[0] :: FloatTensor[batch, 1]`
		
		Note that `len(top_paths) == len(log_probs) == top_paths`
	"""
	batch = tf.shape(logits)[0] # type: ignore
	maxlen = tf.shape(logits)[1] # type: ignore
	time_major_logits = tf.transpose(logits, perm=[1, 0, 2])
	top_paths, log_probs = tf.nn.ctc_greedy_decoder(
		inputs=time_major_logits,
		sequence_length=tf.repeat(maxlen, batch),
		blank_index=0
	)
	return top_paths, log_probs


from hypothesis import given, strategies as st, settings
import tensorflow as tf
from tf.ctc import beam_decode, greedy_decode, onehot_logits

def _test_exact(lab: list[int], decode):
  """Logits generated with `onehot_logits` should trivially decode to `lab`"""
  vlab = tf.sparse.from_dense([lab])
  logits = onehot_logits(vlab)
  paths, _ = decode(logits)
  dense_lab = tf.cast(tf.sparse.to_dense(vlab), tf.int32)
  dense_pred = tf.cast(tf.sparse.to_dense(paths[0]), tf.int32)
  assert tf.reduce_all(tf.equal(dense_lab, dense_pred))

@settings(deadline=None)
@given(st.lists(st.integers(min_value=1, max_value=64), min_size=1, max_size=64))
def test_exact_beam(lab: list[int]):
  _test_exact(lab, beam_decode)

@settings(deadline=None)
@given(st.lists(st.integers(min_value=1, max_value=64), min_size=1, max_size=64))
def test_exact_greedy(lab: list[int]):
  _test_exact(lab, greedy_decode)



import tensorflow as tf
from jaxtyping import Int, Float
from .decoding import beam_decode
import tf.tools as tft

def accuracy(
  labs: Int[tf.SparseTensor, "batch maxlen1"],
  logits: Float[tf.Tensor, "batch maxlen2 vocabsize"],
  beam_width: int = 100, k: int = 1, blank_zero: bool = True
) -> float:
  """Top `k`-accuracy of the batch
  
  (i.e. proportion of samples that correctly predict the corresponding label in one of the top `k` paths)
  - `blank_zero`: if `True` (the default), the blank index is set to 0 (see `tf_ctc.beam_decode` for details)
  """
  top_paths, _ = beam_decode(logits, beam_width=beam_width, top_paths=k, blank_zero=blank_zero)
  labs64 = tf.cast(labs, tf.int64) # thank you tensorflow for always knowing when to enforce precision! xd
  sample_eqs = tft.sparse.any_equal(labs64, top_paths)
  return tf.reduce_mean(tf.cast(sample_eqs, tf.float32))

def edit_distance(
  labs: Int[tf.SparseTensor, "batch maxlen1"],
  logits: Float[tf.Tensor, "batch maxlen2 vocabsize"],
  beam_width: int = 100, k: int = 1, blank_zero: bool = True
) -> float:
  """#### Top `k` edit distance of the batch
  
  (i.e. lowest edit distance between the label and each of the top `k` predictions)
  - `blank_zero`: if `True` (the default), the blank index is set to 0 (see `tf_ctc.beam_decode` for details)
  """
  top_paths, _ = beam_decode(logits, beam_width=beam_width, top_paths=k, blank_zero=blank_zero)
  labs64 = tf.cast(labs, tf.int64) # thank you tensorflow for always knowing when to enforce precision! xd
  sample_eds = tf.reduce_min([tf.edit_distance(labs64, p) for p in top_paths], axis=0)
  return tf.reduce_mean(sample_eds)


import tensorflow as tf
from jaxtyping import Int, Float
from tf.tools import tf_function, pad_dim_to

@tf_function
def onehot_logits_sample(label: Int[tf.Tensor, "maxlen"], value: float = 1e9, blank_index: int = 0, vocabsize: int | None = None, num: int | None = None) -> Float[tf.Tensor, "2*maxlen vocabsize"]:
  """Logits that approach a CTC loss of 0 as `value` grows"""
  depth = tf.cast(vocabsize, tf.int32) # you'll have to ask tensorflow why the heck int64 indices are not valid for tf.one_hot...
  return value*tf.concat([
    [tf.one_hot(x, depth), tf.one_hot(blank_index, depth)]
    for x in tf.unstack(label, num=num)
  ], axis=0) - value # type: ignore
  
@tf_function
def onehot_logits_tf(label: Int[tf.SparseTensor, "batch maxlen"], value: float = 1e9, blank_index: int = 0, *, batch: int, maxlen: int, vocabsize: int = 0) -> Float[tf.Tensor, "batch 2*maxlen vocabsize"]:
  dense = tf.sparse.to_dense(label)
  logits = [onehot_logits_sample(xs, value, blank_index, vocabsize=vocabsize, num=maxlen)[None, ...] for xs in tf.unstack(dense, num=batch)] # type: ignore
  return tf.concat(logits, axis=0) # type: ignore

def onehot_logits(label: Int[tf.SparseTensor, "batch maxlen"], value: float = 1e9, blank_index: int = 0, vocabsize: int | None = None) -> Float[tf.Tensor, "batch 2*maxlen vocabsize"]:
  """Logits that approach a CTC loss of 0 as `value` grows.
  - `label :: Sparse[batch, maxlen]`: a batched sparse label, can be obtained via e.g. `ocr.label('exd4')`
  - `vocabsize`: determines the shape of the logits. Clipped to be greater than `maxlen`
  - Returns `logits :: Tensor[batch, 2*maxlen, vocabsize]`
  """
  batch, maxlen = list(label.dense_shape)
  return onehot_logits_tf(label, value, blank_index=blank_index, batch=int(batch), maxlen=int(maxlen), vocabsize=vocabsize or tf.sparse.reduce_max(label)+1) # type: ignore

def mock_logits(top_preds: list[Int[tf.SparseTensor, "batch maxlen"]], blank_index: int = 0, vocabsize: int | None = None) -> Float[tf.Tensor, "batch 2*maxlen vocabsize"]:
  """Logits that, when CTC-decoded, will roughly follow `top_preds`
  
  E.g:
  ```
  top_preds = R.map(ocr.encode, [['e4', 'Nf6'], ['e5', 'Nd6']])
  z = mock_logits(top_preds)
  paths, _ = ctc.beam_decode(logits, top_paths=2)
  [ocr.decode(p) for p in paths] # [[b'e4', b'Nf6'], [b'e5', b'Nd6']]
  ```
  """
  delta = 1e3
  vocabsize = vocabsize or tf.reduce_max([tf.sparse.reduce_max(l)+1 for l in top_preds]) # type: ignore
  logits = [onehot_logits(l, value=1e9+delta*i, blank_index=blank_index, vocabsize=vocabsize) for i, l in enumerate(reversed(top_preds))]
  maxlen = tf.reduce_max([tf.shape(z)[1] for z in logits]) # type: ignore
  padded_logits = [pad_dim_to(z, axis=1, length=maxlen) for z in logits]
  return tf.reduce_sum(padded_logits, axis=0)